// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package mock

import (
	"fmt"
	"go/ast"
	"go/token"
	"strconv"
	"strings"
	"unicode"
)

const (
	inputFmt     = "arg%d"
	outputFmt    = "ret%d"
	receiverName = "m"
)

// Method represents a method that is being mocked.
type Method struct {
	receiver   Mock
	name       string
	implements *ast.FuncType
}

// MethodFor returns a Method representing typ, using receiver as
// the Method's receiver type and name as the method name.
func MethodFor(receiver Mock, name string, typ *ast.FuncType) Method {
	return Method{
		receiver:   receiver,
		name:       name,
		implements: typ,
	}
}

// Ast returns the ast representation of m.
func (m Method) Ast() *ast.FuncDecl {
	f := &ast.FuncDecl{}
	f.Name = &ast.Ident{Name: m.name}
	f.Type = m.mockType()
	f.Recv = m.recv()
	f.Body = m.body()
	return f
}

// Fields returns the fields that need to be a part of the receiver
// struct for this method.
func (m Method) Fields() []*ast.Field {
	fields := []*ast.Field{
		{
			Names: []*ast.Ident{{Name: m.name + "Called"}},
			Type: &ast.ChanType{
				Dir:   ast.SEND | ast.RECV,
				Value: &ast.Ident{Name: "bool"},
			},
		},
	}
	if params := m.params(false); len(params) > 0 {
		fields = append(fields, &ast.Field{
			Names: []*ast.Ident{{Name: m.name + "Input"}},
			Type:  m.chanStruct(params),
		})
	}
	if results := m.results(false); len(results) > 0 {
		fields = append(fields, &ast.Field{
			Names: []*ast.Ident{{Name: m.name + "Output"}},
			Type:  m.chanStruct(results),
		})
	}
	return fields
}

func (m Method) chanStruct(list []*ast.Field) *ast.StructType {
	typ := &ast.StructType{Fields: &ast.FieldList{}}
	for _, f := range list {
		chanValType := f.Type
		switch src := chanValType.(type) {
		case *ast.ChanType:
			// Receive-only channels require parens, and it seems unfair to leave
			// out send-only channels.
			switch src.Dir {
			case ast.SEND, ast.RECV:
				chanValType = &ast.ParenExpr{X: src}
			}
		case *ast.Ellipsis:
			// The actual value of variadic types is a slice
			chanValType = &ast.ArrayType{Elt: src.Elt}
		}
		names := make([]*ast.Ident, 0, len(f.Names))
		for _, name := range f.Names {
			newName := &ast.Ident{}
			*newName = *name
			names = append(names, newName)
			newName.Name = strings.Title(newName.Name)
		}
		typ.Fields.List = append(typ.Fields.List, &ast.Field{
			Names: names,
			Type: &ast.ChanType{
				Dir:   ast.SEND | ast.RECV,
				Value: chanValType,
			},
		})
	}
	return typ
}

func (m Method) paramChanInit(chanSize int) []ast.Stmt {
	return m.typeChanInit(m.name+"Input", m.params(false), chanSize)
}

func (m Method) returnChanInit(chanSize int) []ast.Stmt {
	return m.typeChanInit(m.name+"Output", m.results(false), chanSize)
}

func (m Method) typeChanInit(fieldName string, fields []*ast.Field, chanSize int) (inputInits []ast.Stmt) {
	for _, field := range fields {
		for _, name := range field.Names {
			inputInits = append(inputInits, &ast.AssignStmt{
				Lhs: []ast.Expr{selectors("m", fieldName, strings.Title(name.String()))},
				Tok: token.ASSIGN,
				Rhs: []ast.Expr{m.makeChan(field.Type, chanSize)},
			})
		}
	}
	return inputInits
}

func (m Method) makeChan(typ ast.Expr, size int) *ast.CallExpr {
	switch src := typ.(type) {
	case *ast.ChanType:
		switch src.Dir {
		case ast.SEND, ast.RECV:
			typ = &ast.ParenExpr{X: src}
		}
	case *ast.Ellipsis:
		// The actual value of variadic types is a slice
		typ = &ast.ArrayType{Elt: src.Elt}
	}
	return &ast.CallExpr{
		Fun: &ast.Ident{Name: "make"},
		Args: []ast.Expr{
			&ast.ChanType{Dir: ast.SEND | ast.RECV, Value: typ},
			&ast.BasicLit{Kind: token.INT, Value: strconv.Itoa(size)},
		},
	}
}

func (m Method) chanInit(chanSize int) []ast.Stmt {
	stmts := []ast.Stmt{
		&ast.AssignStmt{
			Lhs: []ast.Expr{selectors("m", m.name+"Called")},
			Tok: token.ASSIGN,
			Rhs: []ast.Expr{m.makeChan(&ast.Ident{Name: "bool"}, chanSize)},
		},
	}
	stmts = append(stmts, m.paramChanInit(chanSize)...)
	stmts = append(stmts, m.returnChanInit(chanSize)...)
	return stmts
}

func (m Method) recv() *ast.FieldList {
	return &ast.FieldList{
		List: []*ast.Field{
			{
				Names: []*ast.Ident{{Name: receiverName}},
				Type: &ast.StarExpr{
					X: m.receiver.typeExpr(),
				},
			},
		},
	}
}

func (m Method) mockType() *ast.FuncType {
	newTyp := &ast.FuncType{}
	if m.implements.Params != nil {
		newTyp.Params = &ast.FieldList{
			List: m.params(true),
		}
	}
	if m.implements.Results != nil {
		newTyp.Results = &ast.FieldList{
			List: m.results(true),
		}
	}
	return newTyp
}

func (m Method) sendOn(receiver string, fields ...string) *ast.SendStmt {
	return &ast.SendStmt{Chan: selectors(receiver, fields...)}
}

func (m Method) called() ast.Stmt {
	stmt := m.sendOn(receiverName, m.name+"Called")
	stmt.Value = &ast.Ident{Name: "true"}
	return stmt
}

func mockField(idx int, f *ast.Field, fieldFmt string, avoidColl bool) *ast.Field {
	if len(f.Names) == 1 && f.Names[0].Name == "_" {
		f.Names = nil
	}
	if len(f.Names) == 0 {
		if idx < 0 {
			return f
		}
		// Edit the field directly to ensure the same name is used in the mock
		// struct.
		f.Names = []*ast.Ident{{Name: fmt.Sprintf(fieldFmt, idx)}}
		return f
	}

	if !avoidColl {
		return f
	}

	// Here, we want a copy, so that we can use altered names without affecting
	// field names in the mock struct.
	newField := &ast.Field{Type: f.Type}
	for _, n := range f.Names {
		name := n.Name
		if name == receiverName {
			name += "_"
		}
		newField.Names = append(newField.Names, &ast.Ident{Name: name})
	}
	return newField
}

func (m Method) params(avoidColl bool) []*ast.Field {
	var params []*ast.Field
	for idx, f := range m.implements.Params.List {
		params = append(params, mockField(idx, f, inputFmt, avoidColl))
	}
	return params
}

func (m Method) results(avoidColl bool) []*ast.Field {
	if m.implements.Results == nil {
		return nil
	}
	var fields []*ast.Field
	for idx, f := range m.implements.Results.List {
		fields = append(fields, mockField(idx, f, outputFmt, avoidColl))
	}
	return fields
}

func (m Method) inputs() (stmts []ast.Stmt) {
	for _, input := range m.params(true) {
		for _, n := range input.Names {
			// Undo our hack to avoid name collisions with the receiver.
			name := n.Name
			if name == receiverName+"_" {
				name = receiverName
			}
			stmt := m.sendOn(receiverName, m.name+"Input", strings.Title(name))
			stmt.Value = &ast.Ident{Name: n.Name}
			stmts = append(stmts, stmt)
		}
	}
	return stmts
}

// PrependLocalPackage prepends name as the package name for local types
// in m's signature.  This is most often used when mocking types that are
// imported by the local package.
func (m Method) PrependLocalPackage(name string) error {
	if err := m.prependPackage(name, m.implements.Results); err != nil {
		return fmt.Errorf("error adding package selector to results: %w", err)
	}
	if err := m.prependPackage(name, m.implements.Params); err != nil {
		return fmt.Errorf("error adding package selector to params: %w", err)
	}
	return nil
}

func (m Method) prependPackage(name string, fields *ast.FieldList) error {
	if fields == nil {
		return nil
	}
	for _, field := range fields.List {
		typ, err := m.prependTypePackage(name, field.Type)
		if err != nil {
			return fmt.Errorf("could not add package selector to type %v: %w", field.Type, err)
		}
		field.Type = typ
	}
	return nil
}

func (m Method) prependTypePackage(name string, typ ast.Expr) (ast.Expr, error) {
	switch src := typ.(type) {
	case *ast.Ident:
		if !unicode.IsUpper(rune(src.String()[0])) {
			switch src.String() {
			case "bool",
				"string",
				"int", "int8", "int16", "int32", "int64",
				"uint", "uint8", "uint16", "uint32", "uint64", "uintptr",
				"byte",
				"rune",
				"float32", "float64",
				"complex64", "complex128",
				"error", "any":
				return src, nil
			default:
				return nil, fmt.Errorf("cannot add package selector to type %v: %w", typ, ErrUnexported)
			}
		}
		for _, p := range m.receiver.typeParams {
			for _, n := range p.Names {
				if n.Name == src.Name {
					// This is a generic type name and should not have the package name prefixed.
					return src, nil
				}
			}
		}
		return selectors(name, src.String()), nil
	case *ast.FuncType:
		if err := m.prependPackage(name, src.Params); err != nil {
			return nil, fmt.Errorf("cannot add package selector to function params: %w", err)
		}
		if err := m.prependPackage(name, src.Results); err != nil {
			return nil, fmt.Errorf("cannot add package selector to function results: %w", err)
		}
		return src, nil
	case *ast.ArrayType:
		elt, err := m.prependTypePackage(name, src.Elt)
		if err != nil {
			return nil, err
		}
		src.Elt = elt
		return src, nil
	case *ast.MapType:
		key, err := m.prependTypePackage(name, src.Key)
		if err != nil {
			return nil, err
		}
		src.Key = key
		val, err := m.prependTypePackage(name, src.Value)
		if err != nil {
			return nil, err
		}
		src.Value = val
		return src, nil
	case *ast.StarExpr:
		x, err := m.prependTypePackage(name, src.X)
		if err != nil {
			return nil, err
		}
		src.X = x
		return src, nil
	case *ast.Ellipsis:
		elt, err := m.prependTypePackage(name, src.Elt)
		if err != nil {
			return nil, err
		}
		src.Elt = elt
		return src, nil
	case *ast.ChanType:
		val, err := m.prependTypePackage(name, src.Value)
		if err != nil {
			return nil, err
		}
		src.Value = val
		return src, nil
	default:
		return typ, nil
	}
}

func (m Method) recvFrom(receiver string, fields ...string) *ast.UnaryExpr {
	return &ast.UnaryExpr{Op: token.ARROW, X: selectors(receiver, fields...)}
}

func (m Method) returns() (stmts []ast.Stmt) {
	if m.implements.Results == nil {
		return nil
	}
	names := m.results(true)
	populateArgs := []ast.Expr{
		&ast.SelectorExpr{
			X: &ast.Ident{
				Name: receiverName,
			},
			Sel: &ast.Ident{
				Name: "t",
			},
		},
		&ast.BasicLit{
			Kind:  token.STRING,
			Value: strconv.Quote(m.name),
		},
		&ast.SelectorExpr{
			X: &ast.Ident{
				Name: receiverName,
			},
			Sel: &ast.Ident{
				Name: "timeout",
			},
		},
		&ast.SelectorExpr{
			X: &ast.Ident{
				Name: receiverName,
			},
			Sel: &ast.Ident{
				Name: m.name + "Output",
			},
		},
	}
	for _, ns := range names {
		for _, n := range ns.Names {
			populateArgs = append(populateArgs, &ast.UnaryExpr{
				Op: token.AND,
				X:  n,
			})
		}
	}
	populate := &ast.CallExpr{
		Fun: &ast.SelectorExpr{
			X: &ast.Ident{
				Name: "vegr",
			},
			Sel: &ast.Ident{
				Name: "PopulateReturns",
			},
		},
		Args: populateArgs,
	}
	ret := &ast.ReturnStmt{}
	for _, ns := range names {
		for _, n := range ns.Names {
			ret.Results = append(ret.Results, n)
		}
	}
	return []ast.Stmt{
		&ast.ExprStmt{X: populate},
		ret,
	}
}

func (m Method) body() *ast.BlockStmt {
	stmts := []ast.Stmt{
		&ast.ExprStmt{
			X: &ast.CallExpr{
				Fun: &ast.SelectorExpr{
					X: &ast.SelectorExpr{
						X:   &ast.Ident{Name: receiverName},
						Sel: &ast.Ident{Name: "t"},
					},
					Sel: &ast.Ident{Name: "Helper"},
				},
			},
		},
		m.called(),
	}
	stmts = append(stmts, m.inputs()...)
	if returnStmts := m.returns(); returnStmts != nil {
		stmts = append(stmts, m.returns()...)
	}
	return &ast.BlockStmt{
		List: stmts,
	}
}
