// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package mock_test

import (
	"errors"
	"go/ast"
	"go/format"
	"testing"

	"git.sr.ht/~nelsam/hel/mock"
	"github.com/poy/onpar/expect"
)

func TestMockName(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, "type Foo interface{}")
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	expect(m.Name()).To(equal("mockFoo"))
}

func TestMockTypeUnexportedMethod(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
type Foo interface {
  Foo(foo string) int
  bar(bar int) Foo
}
`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	err = m.PrependLocalPackage("foo")
	expect(errors.Is(err, mock.ErrUnexported)).To(equal(true))
}

func TestMock_PrependLocalPackage_UnexportedTypes(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
  Foo(foo string) int
  Bar(bar int) foo
  Baz()
 }
 `)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	err = m.PrependLocalPackage("foo")
	expect(errors.Is(err, mock.ErrUnexported)).To(equal(true))
}

func TestMockTypeDecl(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
  Foo(foo string) int
  Bar(bar int) Foo
  Baz()
  Bacon(func(Eggs) Eggs) func(Eggs) Eggs
 }
 `)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   Foo chan string
  }
  FooOutput struct {
   Ret0 chan int
  }
  BarCalled chan bool
  BarInput struct {
   Bar chan int
  }
  BarOutput struct {
   Ret0 chan Foo
  }
  BazCalled chan bool
  BaconCalled chan bool
  BaconInput struct {
    Arg0 chan func(Eggs) Eggs
  }
  BaconOutput struct {
    Ret0 chan func(Eggs) Eggs
  }
 }
 `))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{m.Decl()}, nil)
	expect(src).To(equal(string(expected)))

	m.PrependLocalPackage("foo")

	expected, err = format.Source([]byte(`
 package foo

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   Foo chan string
  }
  FooOutput struct {
   Ret0 chan int
  }
  BarCalled chan bool
  BarInput struct {
   Bar chan int
  }
  BarOutput struct {
   Ret0 chan foo.Foo
  }
  BazCalled chan bool
  BaconCalled chan bool
  BaconInput struct {
    Arg0 chan func(foo.Eggs) foo.Eggs
  }
  BaconOutput struct {
    Ret0 chan func(foo.Eggs) foo.Eggs
  }
 }
 `))
	expect(err).To(not(haveOccurred()))

	src = source(expect, "foo", []ast.Decl{m.Decl()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockTypeDecl_DirectionalChansGetParens(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
  Foo(foo chan<- int) <-chan int
 }
 `)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   Foo chan (chan<- int)
  }
  FooOutput struct {
   Ret0 chan (<-chan int)
  }
 }
 `))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{m.Decl()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockTypeDecl_VariadicMethods(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
type Foo interface {
 Foo(foo ...int)
}
`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   Foo chan []int
  }
 }
 `))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{m.Decl()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockTypeDecl_ParamsWithoutTypes(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
type Foo interface {
 Foo(foo, bar string)
}
`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
package foo

type mockFoo struct {
 t vegr.T
 timeout time.Duration
 FooCalled chan bool
 FooInput struct {
  Foo, Bar chan string
 }
}
`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{m.Decl()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockConstructor(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
  Foo(foo string) int
  Bar(bar int) string
 }
 `)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 func newMockFoo(t vegr.T, timeout time.Duration) *mockFoo {
  m := &mockFoo{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 300)
  m.FooInput.Foo = make(chan string, 300)
  m.FooOutput.Ret0 = make(chan int, 300)
  m.BarCalled = make(chan bool, 300)
  m.BarInput.Bar = make(chan int, 300)
  m.BarOutput.Ret0 = make(chan string, 300)
  return m
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{m.Constructor(300)}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockConstructor_DirectionalChansGetParens(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
  Foo(foo chan<- int) <-chan int
 }
 `)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 func newMockFoo(t vegr.T, timeout time.Duration) *mockFoo {
  m := &mockFoo{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 200)
  m.FooInput.Foo = make(chan (chan<- int), 200)
  m.FooOutput.Ret0 = make(chan (<-chan int), 200)
  return m
 }
 `))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{m.Constructor(200)}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockConstructor_VariadicParams(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
  Foo(foo ...int)
 }
 `)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 func newMockFoo(t vegr.T, timeout time.Duration) *mockFoo {
  m := &mockFoo{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 200)
  m.FooInput.Foo = make(chan []int, 200)
  return m
 }
 `))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{m.Constructor(200)}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockGenerics(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo[T any] interface {
  Foo(foo T) T
 }
 `)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 type mockFoo[T any] struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   Foo chan T
  }
  FooOutput struct {
   Ret0 chan T
  }
 }
 `))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{m.Decl()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockAst(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
  Bar(bar string)
  Baz() (baz int)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	decls := m.Ast(300)
	expect(decls).To(haveLen(4))
	expect(decls[0]).To(equal(m.Decl()))
	expect(decls[1]).To(equal(m.Constructor(300)))
	expect(m.Methods()).To(haveLen(2))
	expect(decls[2]).To(equal(m.Methods()[0].Ast()))
	expect(decls[3]).To(equal(m.Methods()[1].Ast()))
}
