// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package pers_test

import (
	"fmt"
	"reflect"
	"testing"
	"time"

	"git.sr.ht/~nelsam/hel/pkg/pers"
	"github.com/poy/onpar"
	"github.com/poy/onpar/expect"
)

type mockT struct {
	fatalled bool
	fatalMsg string
	cleanup  func()
}

func newMockT() *mockT {
	return &mockT{}
}

func (t *mockT) Fatalf(v string, args ...interface{}) {
	t.fatalled = true
	t.fatalMsg = fmt.Sprintf(v, args...)
}

func (t *mockT) Cleanup(f func()) {
	t.cleanup = f
}

func TestConsistentlyReturn(t *testing.T) {
	o := onpar.BeforeEach(onpar.New(t), func(t *testing.T) expectation {
		return expect.New(t)
	})
	defer o.Run()

	for _, test := range []struct {
		name     string
		returnTo interface{}
		values   []interface{}
		panicMsg string
	}{
		{
			name:     "an empty struct",
			returnTo: struct{}{},
			panicMsg: "pers: consistently returning [] on mock (struct {}) is not possible: sel: empty struct (struct {}) is not a valid mock field",
		},
		{
			name:     "an invalid struct",
			returnTo: struct{ Foo int }{},
			values:   []interface{}{1},
			panicMsg: "pers: consistently returning [1] on mock (struct { Foo int }) is not possible: sel: field 0 of type struct { Foo int } cannot be used as a mock field: sel: type int is neither a struct of channels nor a channel",
		},
		{
			name: "too few arguments",
			returnTo: struct {
				Foo chan int
				Bar chan string
			}{
				Foo: make(chan int, 1),
				Bar: make(chan string, 1),
			},
			values:   []interface{}{1},
			panicMsg: "pers: consistently returning [1] on mock (struct { Foo chan int; Bar chan string }) is not possible: sel: expected 2 arguments for mock (struct { Foo chan int; Bar chan string }); got 1",
		},
		{
			name:     "too many arguments",
			returnTo: make(chan int, 1),
			values:   []interface{}{2, "foo"},
			panicMsg: "pers: consistently returning [2 foo] on mock (chan int) is not possible: sel: expected 1 argument for mock (chan int); got 2",
		},
	} {
		test := test
		o.Spec(fmt.Sprintf("it panics with %v", test.name), func(expect expectation) {
			mT := newMockT()
			defer func() {
				r := recover()
				expect(r).To(not(beNil()))
				expect(fmt.Sprintf("%v", r)).To(equal(test.panicMsg))
			}()
			pers.ConsistentlyReturn(mT, test.returnTo, test.values...)
		})
	}

	o.Spec("it handles nil values correctly", func(expect expectation) {
		mT := newMockT()
		c := make(chan error)
		stop := pers.ConsistentlyReturn(mT, c, nil)
		expect(mT.fatalled).To(not(beTrue()))
		defer stop()
		for i := 0; i < 1000; i++ {
			expect(<-c).To(equal(nil))
		}
	})

	o.Spec("it consistently returns on a channel", func(expect expectation) {
		mT := newMockT()
		c := make(chan int)
		stop := pers.ConsistentlyReturn(mT, c, 1)
		expect(mT.fatalled).To(not(beTrue()))
		defer stop()
		for i := 0; i < 1000; i++ {
			expect(<-c).To(equal(1))
		}
	})

	o.Spec("it consistently returns on a struct of channels", func(expect expectation) {
		mT := newMockT()
		type fooReturns struct {
			Foo chan string
			Bar chan bool
		}
		v := fooReturns{make(chan string), make(chan bool)}
		stop := pers.ConsistentlyReturn(mT, v, "foo", true)
		expect(mT.fatalled).To(not(beTrue()))
		defer stop()
		for i := 0; i < 1000; i++ {
			expect(<-v.Foo).To(equal("foo"))
			expect(<-v.Bar).To(equal(true))
		}
	})

	o.Spec("it stops returning after stop is called", func(expect expectation) {
		mT := newMockT()
		c := make(chan string)
		stop := pers.ConsistentlyReturn(mT, c, "foo")
		expect(mT.fatalled).To(not(beTrue()))
		stop()
		expect(c).To(not(receiveMatcher{timeout: 100 * time.Millisecond}))
	})

	o.Spec("it calls T.Cleanup with the stop function", func(expect expectation) {
		mT := newMockT()
		c := make(chan string)
		stop := pers.ConsistentlyReturn(mT, c, "foo")
		expect(mT.fatalled).To(not(beTrue()))
		ptrFmt := "%p"
		stopAddr := fmt.Sprintf(ptrFmt, stop)
		cleanupAddr := fmt.Sprintf(ptrFmt, mT.cleanup)
		expect(stopAddr).To(startWith("0x")) // just double checking that this actually printed as an address
		expect(stopAddr).To(equal(cleanupAddr))
		stop()
	})

	o.Spec("stop and cleanup can both be called successfully", func(expect expectation) {
		mT := newMockT()
		c := make(chan string)
		stop := pers.ConsistentlyReturn(mT, c, "foo")
		expect(mT.fatalled).To(not(beTrue()))
		stop()
		mT.cleanup()
	})
}

type receiveMatcher struct {
	timeout time.Duration
}

func (r receiveMatcher) Match(actual interface{}) (interface{}, error) {
	cases := []reflect.SelectCase{
		{Dir: reflect.SelectRecv, Chan: reflect.ValueOf(actual)},
		{Dir: reflect.SelectRecv, Chan: reflect.ValueOf(time.After(r.timeout))},
	}
	i, v, _ := reflect.Select(cases)
	if i == 1 {
		return actual, fmt.Errorf("timed out after %s waiting for %#v to receive", r.timeout, actual)
	}
	return v.Interface(), nil
}
