// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package typ_test

import (
	"go/ast"
	"testing"
	"time"

	"git.sr.ht/~nelsam/hel/pkg/pers"
	"git.sr.ht/~nelsam/hel/typ"
	"github.com/poy/onpar"
	"github.com/poy/onpar/expect"
	"github.com/poy/onpar/matchers"
	"golang.org/x/tools/go/packages"
)

type expectation = expect.Expectation

var (
	equal        = matchers.Equal
	not          = matchers.Not
	haveOccurred = matchers.HaveOccurred
	haveLen      = matchers.HaveLen
	beNil        = matchers.BeNil
	beTrue       = matchers.BeTrue
)

func TestTypes(t *testing.T) {
	type testCtx struct {
		t      *testing.T
		expect expectation
		dir    *mockGoDir
	}

	o := onpar.BeforeEach(onpar.New(t), func(t *testing.T) testCtx {
		return testCtx{t: t, expect: expect.New(t), dir: newMockGoDir(t, time.Second)}
	})
	defer o.Run()

	o.Spec("Load_EmptyInterface", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "foo",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, "type Foo interface {}"),
			},
		})
		found := typ.Load(false, tt.dir)
		tt.expect(found.Slice()).To(haveLen(1))
		tt.expect(found.Slice()[0].Len()).To(equal(1))
		tt.expect(found.Slice()[0].Dir()).To(equal("/some/path"))
		tt.expect(found.Slice()[0].Package()).To(equal("foo"))
	})

	o.Spec("Filter", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "foo",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Foo interface {}
    type Bar interface {}
    type FooBar interface {}
    type BarFoo interface {}
    `),
			},
		})
		found := typ.Load(false, tt.dir)
		tt.expect(found.Slice()).To(haveLen(1))
		tt.expect(found.Slice()[0].Len()).To(equal(4))

		notFiltered := found.Filter()
		tt.expect(notFiltered.Slice()).To(haveLen(1))
		tt.expect(notFiltered.Slice()[0].Len()).To(equal(4))

		foos := found.Filter("Foo")
		tt.expect(foos.Slice()).To(haveLen(1))
		tt.expect(foos.Slice()[0].Len()).To(equal(1))
		tt.expect(typeSpec(tt.t, tt.expect, foos.Slice()[0].Decls()[0]).Name.String()).To(equal("Foo"))

		fooPrefixes := found.Filter("Foo.*")
		tt.expect(fooPrefixes.Slice()).To(haveLen(1))
		tt.expect(fooPrefixes.Slice()[0].Len()).To(equal(2))
		expectNamesToMatch(tt.t, tt.expect, fooPrefixes.Slice()[0].Decls(), "Foo", "FooBar")

		fooPostfixes := found.Filter(".*Foo")
		tt.expect(fooPostfixes.Slice()).To(haveLen(1))
		tt.expect(fooPostfixes.Slice()[0].Len()).To(equal(2))
		expectNamesToMatch(tt.t, tt.expect, fooPostfixes.Slice()[0].Decls(), "Foo", "BarFoo")

		fooContainers := found.Filter("Foo.*", ".*Foo")
		tt.expect(fooContainers.Slice()).To(haveLen(1))
		tt.expect(fooContainers.Slice()[0].Len()).To(equal(3))
		expectNamesToMatch(tt.t, tt.expect, fooContainers.Slice()[0].Decls(), "Foo", "FooBar", "BarFoo")
	})

	o.Spec("LocalDependencies", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    type Bar interface{
        Bar(Foo) Foo
    }`),
				parse(tt.t, tt.expect, `

    type Foo interface {
        Foo()
    }`),
			},
		})

		found := typ.Load(false, tt.dir)

		tt.expect(found.Slice()).To(haveLen(1))
		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(2))

		var foo, bar *ast.TypeSpec
		for _, m := range mockables {
			mockable := typeSpec(tt.t, tt.expect, m)
			switch mockable.Name.String() {
			case "Bar":
				bar = mockable
			case "Foo":
				foo = mockable
			}
		}
		if bar == nil {
			t.Fatal("expected to find a Bar type")
		}

		tt.expect(found.Slice()).To(haveLen(1))
		dependencies := found.Slice()[0].Dependencies(bar.Name.Name)
		tt.expect(dependencies).To(haveLen(1))
		tt.expect(dependencies[0].Type()).To(equal(foo.Type))
		tt.expect(dependencies[0].PkgName()).To(equal(""))
		tt.expect(dependencies[0].PkgPath()).To(equal(""))
	})

	o.Group("concrete dependencies", func() {
		o.Spec("it ignores return types of concrete dependencies", func(tt testCtx) {
			pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
			pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
				Name: "bar",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `

		    func Bar(int) Foo {
		        return nil
		    }`),
					parse(tt.t, tt.expect, `

		    type Foo interface {
		        Foo()
		    }`),
				},
			})

			found := typ.Load(false, tt.dir)

			tt.expect(found.Slice()).To(haveLen(1))
			mockables := found.Slice()[0].Decls()
			tt.expect(mockables).To(haveLen(2))

			var (
				bar *ast.FuncDecl
			)
			for _, m := range mockables {
				switch d := m.(type) {
				case *ast.FuncDecl:
					bar = d
				}
			}
			if bar == nil {
				t.Fatal("expected to find a Bar function")
			}

			dependencies := found.Slice()[0].Dependencies(bar.Name.Name)
			tt.expect(dependencies).To(haveLen(0))
		})

		o.Spec("it understands dependencies of function parameters", func(tt testCtx) {
			pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
			pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
				Name: "bar",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `

		    func Bar(f Foo) int {
		        return -1
		    }`),
					parse(tt.t, tt.expect, `

		    type Foo interface {
		        Foo()
		    }`),
				},
			})

			found := typ.Load(false, tt.dir)

			tt.expect(found.Slice()).To(haveLen(1))
			mockables := found.Slice()[0].Decls()
			tt.expect(mockables).To(haveLen(2))

			var (
				foo *ast.TypeSpec
				bar *ast.FuncDecl
			)
			for _, m := range mockables {
				switch d := m.(type) {
				case *ast.GenDecl:
					foo = typeSpec(tt.t, tt.expect, m)
				case *ast.FuncDecl:
					bar = d
				}
			}
			if bar == nil {
				t.Fatal("expected to find a Bar function")
			}

			dependencies := found.Slice()[0].Dependencies(bar.Name.Name)
			tt.expect(dependencies).To(haveLen(1))
			tt.expect(dependencies[0].Type()).To(equal(foo.Type))
			tt.expect(dependencies[0].PkgName()).To(equal(""))
			tt.expect(dependencies[0].PkgPath()).To(equal(""))
		})

		o.Spec("it understands the dependencies of method parameters", func(tt testCtx) {
			pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
			pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
				Name: "bar",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `

            type Bar struct {}

		    func (b Bar) Baz(f Foo) int {
		        return -1
		    }

            // since bacon is unexported, it should not result in new
            // dependencies without unexported == true.
            func (b Bar) bacon(e Eggs) int {
                return -1
            }`),
					parse(tt.t, tt.expect, `

		    type Foo interface {
		        Foo()
		    }

            type Eggs interface {
                Eggs()
            }`),
				},
			})

			found := typ.Load(false, tt.dir)

			tt.expect(found.Slice()).To(haveLen(1))
			mockables := found.Slice()[0].Decls()
			tt.expect(mockables).To(haveLen(5))

			var (
				foo *ast.TypeSpec
				baz *ast.FuncDecl
			)
			for _, m := range mockables {
				switch d := m.(type) {
				case *ast.GenDecl:
					if d.Specs[0].(*ast.TypeSpec).Name.Name != "Foo" {
						continue
					}
					foo = typeSpec(tt.t, tt.expect, m)
				case *ast.FuncDecl:
					if d.Name.Name != "Baz" {
						continue
					}
					baz = d
				}
			}
			if baz == nil {
				t.Fatal("expected to find a Baz function")
			}

			// The method name should not have dependencies
			tt.expect(found.Slice()[0].Dependencies(baz.Name.Name)).To(haveLen(0))

			tt.expect(baz.Recv).To(not(beNil()))
			tt.expect(baz.Recv.List).To(haveLen(1))

			dependencies := found.Slice()[0].Dependencies(baz.Recv.List[0].Type.(*ast.Ident).Name)
			tt.expect(dependencies).To(haveLen(1))
			tt.expect(dependencies[0].Type()).To(equal(foo.Type))
			tt.expect(dependencies[0].PkgName()).To(equal(""))
			tt.expect(dependencies[0].PkgPath()).To(equal(""))
		})

		o.Spec("it understands the dependencies of struct fields", func(tt testCtx) {
			pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
			pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
				Name: "bar",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `

            type Bar struct {
                Foo Foo

                // eggs will be ignored since with unexported==false
                eggs Eggs
            }`),
					parse(tt.t, tt.expect, `

		    type Foo interface {
		        Foo()
		    }

            type Eggs interface {
                Eggs()
            }`),
				},
			})

			found := typ.Load(false, tt.dir)

			tt.expect(found.Slice()).To(haveLen(1))
			mockables := found.Slice()[0].Decls()
			tt.expect(mockables).To(haveLen(3))

			var (
				foo *ast.TypeSpec
				bar *ast.TypeSpec
			)
			for _, m := range mockables {
				mockable := typeSpec(tt.t, tt.expect, m)
				switch mockable.Name.String() {
				case "Bar":
					bar = mockable
				case "Foo":
					foo = mockable
				}
			}
			if bar == nil {
				t.Fatal("expected to find a Bar type")
			}

			dependencies := found.Slice()[0].Dependencies(bar.Name.Name)
			tt.expect(dependencies).To(haveLen(1))
			tt.expect(dependencies[0].Type()).To(equal(foo.Type))
			tt.expect(dependencies[0].PkgName()).To(equal(""))
			tt.expect(dependencies[0].PkgPath()).To(equal(""))
		})
	})

	o.Spec("it can mock unexported fields and methods", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

            type Bar struct {
                foo Foo
            }

            func (b Bar) eggs(Eggs) {
            }`),
				parse(tt.t, tt.expect, `

		    type Foo interface {
		        Foo()
		    }

            type Eggs interface {
                Eggs()
            }`),
			},
		})

		found := typ.Load(true, tt.dir)

		tt.expect(found.Slice()).To(haveLen(1))
		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(4))

		var (
			foo  *ast.TypeSpec
			bar  *ast.TypeSpec
			eggs *ast.TypeSpec
		)
		for _, m := range mockables {
			if _, ok := m.(*ast.FuncDecl); ok {
				continue
			}
			mockable := typeSpec(tt.t, tt.expect, m)
			switch mockable.Name.String() {
			case "Bar":
				bar = mockable
			case "Foo":
				foo = mockable
			case "Eggs":
				eggs = mockable
			}
		}
		if bar == nil {
			t.Fatal("expected to find a Bar type")
		}

		dependencies := found.Slice()[0].Dependencies(bar.Name.Name)
		tt.expect(dependencies).To(haveLen(2))
		fooDep, eggsDep := dependencies[0], dependencies[1]
		if fooDep.TypeName() != "Foo" {
			fooDep, eggsDep = eggsDep, fooDep
		}
		tt.expect(fooDep.Type()).To(equal(foo.Type))
		tt.expect(eggsDep.Type()).To(equal(eggs.Type))
	})

	o.Spec("CyclicDependencies", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    type Bar interface{
        Bar(Foo) Foo
    }`),
				parse(tt.t, tt.expect, `

    type Foo interface {
        Foo(Bar) Bar
    }`),
			},
		})

		foundCh := make(chan typ.Dirs)
		go func() {
			// This test checks for an infinite loop. In order for the test to
			// fail quicker, we run the production code in a goroutine and add
			// timeouts to the assertions.
			foundCh <- typ.Load(false, tt.dir)
		}()

		var found typ.Dirs
		select {
		case found = <-foundCh:
		case <-time.After(time.Second):
			tt.t.Fatalf("timed out waiting for typ.Load to return")
		}
		tt.expect(found.Slice()).To(haveLen(1))
		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(2))

		var foo, bar *ast.TypeSpec
		for _, m := range mockables {
			typ := typeSpec(tt.t, tt.expect, m)
			switch typ.Name.String() {
			case "Bar":
				bar = typ
			case "Foo":
				foo = typ
			}
		}
		if bar == nil {
			t.Fatal("expected to find a Bar type")
		}

		barDeps := found.Slice()[0].Dependencies(bar.Name.Name)
		tt.expect(barDeps).To(haveLen(1))
		tt.expect(barDeps[0].Type()).To(equal(foo.Type))
		tt.expect(barDeps[0].PkgName()).To(equal(""))
		tt.expect(barDeps[0].PkgPath()).To(equal(""))

		fooDeps := found.Slice()[0].Dependencies(foo.Name.Name)
		tt.expect(fooDeps).To(haveLen(1))
		tt.expect(fooDeps[0].Type()).To(equal(bar.Type))
		tt.expect(fooDeps[0].PkgName()).To(equal(""))
		tt.expect(fooDeps[0].PkgPath()).To(equal(""))
	})

	o.Spec("ImportedDependencies", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import "some/path/to/foo"

    type Bar interface{
        Bar(foo.Foo) foo.Bar
    }`),
			},
		})

		pkgName := "foo"
		pkg := &packages.Package{
			Name: pkgName,
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Foo interface {
        Foo()
    }

	type Bar interface {
		Bar()
	}`),
			},
		}
		pers.ConsistentlyReturn(tt.t, tt.dir.ImportOutput, pkg, nil)

		found := typ.Load(false, tt.dir)
		tt.expect(found.Slice()).To(haveLen(1))
		tt.expect(tt.dir.ImportCalled).To(haveLen(1))

		tt.expect(<-tt.dir.ImportInput.Path).To(equal("some/path/to/foo"))

		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(2))
		if mockables[1] == nil {
			t.Fatal("expected mockables[0] to be non-nil")
		}

		dependencies := found.Slice()[0].Dependencies(typeSpec(tt.t, tt.expect, mockables[1]).Name.Name)
		tt.expect(dependencies).To(haveLen(2))

		names := make(map[string]bool)
		for _, dependent := range dependencies {
			tt.expect(dependent.PkgName()).To(equal("foo"))
			tt.expect(dependent.PkgPath()).To(equal("some/path/to/foo"))
			names[dependent.TypeName()] = true
		}
		tt.expect(names).To(equal(map[string]bool{"Foo": true, "Bar": true}))
	})

	o.Spec("AliasedImportedDependencies", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import baz "some/path/to/foo"

    type Bar interface{
        Bar(baz.Foo) baz.Bar
    }`),
			},
		})

		pkgName := "foo"
		pkg := &packages.Package{
			Name: pkgName,
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Foo interface {
        Foo()
    }

	type Bar interface {
		Bar()
	}`),
			},
		}
		pers.ConsistentlyReturn(tt.t, tt.dir.ImportOutput, pkg, nil)

		found := typ.Load(false, tt.dir)
		tt.expect(tt.dir.ImportCalled).To(haveLen(1))
		tt.expect(<-tt.dir.ImportInput.Path).To(equal("some/path/to/foo"))

		tt.expect(found.Slice()).To(haveLen(1))
		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(2))

		dependencies := found.Slice()[0].Dependencies(typeSpec(tt.t, tt.expect, mockables[1]).Name.Name)
		tt.expect(dependencies).To(haveLen(2))

		names := make(map[string]bool)
		for _, dependent := range dependencies {
			tt.expect(dependent.PkgName()).To(equal("baz"))
			tt.expect(dependent.PkgPath()).To(equal("some/path/to/foo"))
			names[dependent.TypeName()] = true
		}
		tt.expect(names).To(equal(map[string]bool{"Foo": true, "Bar": true}))
	})

	o.Spec("NestedAliasedImportedDependencies", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import "some/path/to/foo"

    type Bar interface{
        Bar(foo.Foo)
    }`),
			},
		})

		foundCh := make(chan typ.Dirs)
		go func() {
			foundCh <- typ.Load(false, tt.dir)
		}()

		tt.expect(tt.dir).To(pers.HaveMethodExecuted("Import",
			pers.Within(time.Second),
			pers.WithArgs("some/path/to/foo"),
			pers.Returning(&packages.Package{
				Name: "foo",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `
    import bazinga "a/dir/for/baz"

    type Foo interface{
        Foo(bazinga.Baz)
    }
    `),
				},
			}, nil),
		))

		tt.expect(tt.dir).To(pers.HaveMethodExecuted("Import",
			pers.Within(time.Second),
			pers.WithArgs("a/dir/for/baz"),
			pers.Returning(&packages.Package{
				Name: "baz",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `
    type Baz interface {
        Baz()
    }`),
				},
			}, nil),
		))

		found := <-foundCh

		tt.expect(found.Slice()).To(haveLen(1))
		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(2))

		dependencies := found.Slice()[0].Dependencies(typeSpec(tt.t, tt.expect, mockables[1]).Name.Name)
		tt.expect(dependencies).To(haveLen(2))

		var baz typ.Dependency
		for _, dependent := range dependencies {
			switch dependent.TypeName() {
			case "Foo":
			case "Baz":
				baz = dependent
			default:
				tt.t.Fatalf("got unexpected type name %v", dependent.TypeName())
			}
		}
		tt.expect(baz.PkgName()).To(equal("bazinga"))
		tt.expect(baz.PkgPath()).To(equal("a/dir/for/baz"))
		tt.expect(baz.TypeName()).To(equal("Baz"))
	})

	o.Spec("ImportedChannelDependencies", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import "some/path/to/foo"

    type Bar interface{
        Bar(foo.Foo)
    }`),
			},
		})

		foundCh := make(chan typ.Dirs)
		go func() {
			foundCh <- typ.Load(false, tt.dir)
		}()

		tt.expect(tt.dir).To(pers.HaveMethodExecuted("Import",
			pers.Within(time.Second),
			pers.WithArgs("some/path/to/foo"),
			pers.Returning(&packages.Package{
				Name: "foo",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `
    import bazinga "a/dir/for/baz"

    type Foo interface{
        Foo() chan Bar
        Baz() chan bazinga.Baz
    }

    type Bar string
    `),
				},
			}, nil),
		))

		tt.expect(tt.dir).To(pers.HaveMethodExecuted("Import",
			pers.Within(time.Second),
			pers.WithArgs("a/dir/for/baz"),
			pers.Returning(&packages.Package{
				Name: "baz",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `
    type Baz string`),
				},
			}, nil),
		))

		found := <-foundCh

		tt.expect(found.Slice()).To(haveLen(1))
		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(2))

		dependencies := found.Slice()[0].Dependencies(typeSpec(tt.t, tt.expect, mockables[1]).Name.Name)
		tt.expect(dependencies).To(haveLen(2))

		var foo, baz typ.Dependency
		for _, dependent := range dependencies {
			switch dependent.TypeName() {
			case "Foo":
				foo = dependent
			case "Baz":
				baz = dependent
			default:
				tt.t.Fatalf("got unexpected type name %v", dependent.TypeName())
			}
		}
		tt.expect(foo.PkgName()).To(equal("foo"))
		tt.expect(foo.PkgPath()).To(equal("some/path/to/foo"))
		tt.expect(foo.TypeName()).To(equal("Foo"))

		tt.expect(baz.PkgName()).To(equal("bazinga"))
		tt.expect(baz.PkgPath()).To(equal("a/dir/for/baz"))
		tt.expect(baz.TypeName()).To(equal("Baz"))
	})

	o.Spec("AliasedImportedDependencies", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import baz "some/path/to/foo"

    type Bar interface{
        Bar(baz.Foo) baz.Bar
    }`),
			},
		})

		pkgName := "foo"
		pkg := &packages.Package{
			Name: pkgName,
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Foo interface {
        Foo()
    }

	type Bar interface {
		Bar()
	}`),
			},
		}
		pers.ConsistentlyReturn(tt.t, tt.dir.ImportOutput, pkg, nil)

		found := typ.Load(false, tt.dir)
		tt.expect(tt.dir.ImportCalled).To(haveLen(1))
		tt.expect(<-tt.dir.ImportInput.Path).To(equal("some/path/to/foo"))

		tt.expect(found.Slice()).To(haveLen(1))
		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(2))

		dependencies := found.Slice()[0].Dependencies(typeSpec(tt.t, tt.expect, mockables[1]).Name.Name)
		tt.expect(dependencies).To(haveLen(2))

		names := make(map[string]bool)
		for _, dependent := range dependencies {
			tt.expect(dependent.PkgName()).To(equal("baz"))
			tt.expect(dependent.PkgPath()).To(equal("some/path/to/foo"))
			names[dependent.TypeName()] = true
		}
		tt.expect(names).To(equal(map[string]bool{"Foo": true, "Bar": true}))
	})

	o.Spec("NestedImportedDependencies", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import "some/path/to/foo"

    type Fooer interface{
        Foo() foo.Foo
    }`),
			},
		})

		fooName := "foo"
		foo := &packages.Package{
			Name: fooName,
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    import "some/path/to/bar"

    type Foo interface {
        Bar() bar.Bar
    }`),
			},
		}

		barName := "bar"
		bar := &packages.Package{
			Name: barName,
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Bar interface {
        Baz()
    }`),
			},
		}

		dirs := make(chan typ.Dirs)
		go func() {
			dirs <- typ.Load(false, tt.dir)
		}()

		tt.expect(tt.dir).To(
			pers.HaveMethodExecuted("Import",
				pers.Within(time.Second),
				pers.WithArgs("some/path/to/foo"),
			),
		)
		pers.Return(tt.dir.ImportOutput, foo, nil)

		tt.expect(tt.dir).To(
			pers.HaveMethodExecuted("Import",
				pers.Within(time.Second),
				pers.WithArgs("some/path/to/bar"),
			),
		)
		pers.Return(tt.dir.ImportOutput, bar, nil)

		found := <-dirs
		tt.expect(found.Slice()).To(haveLen(1))

		mockables := found.Slice()[0].Decls()
		tt.expect(mockables).To(haveLen(2))
		if mockables[1] == nil {
			t.Fatal("expected mockables[0] to be non-nil")
		}

		dependencies := found.Slice()[0].Dependencies(typeSpec(tt.t, tt.expect, mockables[1]).Name.Name)
		tt.expect(dependencies).To(haveLen(2))
		for _, dependent := range dependencies {
			switch dependent.Name() {
			case "some/path/to/foo.Foo":
				tt.expect(dependent.PkgName()).To(equal("foo"))
				tt.expect(dependent.PkgPath()).To(equal("some/path/to/foo"))
			case "some/path/to/bar.Bar":
				tt.expect(dependent.PkgName()).To(equal("bar"))
				tt.expect(dependent.PkgPath()).To(equal("some/path/to/bar"))
			default:
				tt.t.Fatalf("unexpected dependency %v", dependent.Name())
			}
		}
	})

	// TestAnonymousError is testing the only case (as of go 1.7) where
	// a builtin is an interface type.
	o.Spec("AnonymousError", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "foo",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Foo interface{
        error
    }`),
			},
		})
		found := typ.Load(false, tt.dir)
		tt.expect(found.Slice()).To(haveLen(1))

		typs := found.Slice()[0].Decls()
		tt.expect(typs).To(haveLen(1))

		decl := typs[0]
		tt.expect(decl).To(not(beNil()))

		inter := typeSpec(tt.t, tt.expect, decl).Type.(*ast.InterfaceType)
		tt.expect(inter.Methods.List).To(haveLen(1))
		err := inter.Methods.List[0]
		tt.expect(err.Names[0].String()).To(equal("Error"))
		_, isFunc := err.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())
	})

	o.Spec("AnonymousLocalTypes", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "foo",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Bar interface{
        Foo
        Bar()
    }`),
				parse(tt.t, tt.expect, `
    type Foo interface{
        Foo()
    }`),
			},
		})
		found := typ.Load(false, tt.dir)
		tt.expect(found.Slice()).To(haveLen(1))

		typs := found.Slice()[0].Decls()
		tt.expect(typs).To(haveLen(2))

		spec := find(tt.t, tt.expect, typs, "Bar")
		tt.expect(spec).To(not(beNil()))
		inter := spec.Type.(*ast.InterfaceType)
		tt.expect(inter.Methods.List).To(haveLen(2))
		foo := inter.Methods.List[0]
		tt.expect(foo.Names[0].String()).To(equal("Foo"))
		_, isFunc := foo.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())
	})

	o.Spec("AnonymousImportedTypes", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import "some/path/to/foo"

    type foo = foo.Foo`),
				parse(tt.t, tt.expect, `

    type Bar interface{
        foo
        Bar()
    }`),
			},
		})

		pkgName := "foo"
		pkg := &packages.Package{
			Name: pkgName,
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Foo interface {
        Foo(x X) Y
    }

	type X int
	type Y int`),
			},
		}
		pers.ConsistentlyReturn(tt.t, tt.dir.ImportOutput, pkg, nil)

		found := typ.Load(false, tt.dir)

		// 2 calls: 1 for the initial import, then deps import for X and Y
		tt.expect(tt.dir.ImportCalled).To(haveLen(2))
		tt.expect(<-tt.dir.ImportInput.Path).To(equal("some/path/to/foo"))

		tt.expect(found.Slice()).To(haveLen(1))
		typs := found.Slice()[0].Decls()
		tt.expect(typs).To(haveLen(3))

		decl := find(tt.t, tt.expect, typs, "Bar")
		tt.expect(decl).To(not(beNil()))
		inter := decl.Type.(*ast.InterfaceType)
		tt.expect(inter.Methods.List).To(haveLen(2))

		foo := inter.Methods.List[0]
		tt.expect(foo.Names[0].String()).To(equal("Foo"))
		f, isFunc := foo.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())
		tt.expect(f.Params.List).To(haveLen(1))
		tt.expect(f.Results.List).To(haveLen(1))
		expr, isSelector := f.Params.List[0].Type.(*ast.SelectorExpr)
		tt.expect(isSelector).To(beTrue())
		tt.expect(expr.X.(*ast.Ident).String()).To(equal("foo"))
		tt.expect(expr.Sel.String()).To(equal("X"))
		expr, isSelector = f.Results.List[0].Type.(*ast.SelectorExpr)
		tt.expect(isSelector).To(beTrue())
		tt.expect(expr.X.(*ast.Ident).String()).To(equal("foo"))
		tt.expect(expr.Sel.String()).To(equal("Y"))
	})

	o.Spec("AnonymousAliasedImportedTypes", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import baz "some/path/to/foo"

    type Bar interface{
        baz.Foo
        Bar()
    }`),
			},
		})

		pkgName := "foo"
		pkg := &packages.Package{
			Name: pkgName,
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Foo interface {
        Foo(x X) Y
    }

	type X int
	type Y int`),
			},
		}
		pers.ConsistentlyReturn(tt.t, tt.dir.ImportOutput, pkg, nil)

		found := typ.Load(false, tt.dir)

		// 2 calls: 1 for the initial import, then deps imports for X and Y
		tt.expect(tt.dir.ImportCalled).To(haveLen(2))
		tt.expect(<-tt.dir.ImportInput.Path).To(equal("some/path/to/foo"))

		tt.expect(found.Slice()).To(haveLen(1))
		typs := found.Slice()[0].Decls()
		tt.expect(typs).To(haveLen(2))

		decl := typs[1]
		tt.expect(decl).To(not(beNil()))
		inter := typeSpec(tt.t, tt.expect, decl).Type.(*ast.InterfaceType)
		tt.expect(inter.Methods.List).To(haveLen(2))

		foo := inter.Methods.List[0]
		tt.expect(foo.Names[0].String()).To(equal("Foo"))
		f, isFunc := foo.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())
		tt.expect(f.Params.List).To(haveLen(1))
		tt.expect(f.Results.List).To(haveLen(1))
		expr, isSelector := f.Params.List[0].Type.(*ast.SelectorExpr)
		tt.expect(isSelector).To(beTrue())
		tt.expect(expr.X.(*ast.Ident).String()).To(equal("baz"))
		tt.expect(expr.Sel.String()).To(equal("X"))
		expr, isSelector = f.Results.List[0].Type.(*ast.SelectorExpr)
		tt.expect(isSelector).To(beTrue())
		tt.expect(expr.X.(*ast.Ident).String()).To(equal("baz"))
		tt.expect(expr.Sel.String()).To(equal("Y"))
	})

	o.Spec("AnonymousImportedTypeAliases", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "foo",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import "some/path/to/bar"

    type Foo interface{
        bar.Bar
        Foo()
    }`),
			},
		})

		foundCh := make(chan typ.Dirs)
		go func() {
			foundCh <- typ.Load(false, tt.dir)
		}()

		tt.expect(tt.dir).To(pers.HaveMethodExecuted("Import",
			pers.Within(time.Second),
			pers.WithArgs("some/path/to/bar"),
			pers.Returning(&packages.Package{
				Name: "bar",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `
    import "a/dir/for/baz"

    type Bar = baz.Bar
    `),
				},
			}, nil),
		))

		tt.expect(tt.dir).To(pers.HaveMethodExecuted("Import",
			pers.Within(time.Second),
			pers.WithArgs("a/dir/for/baz"),
			pers.Returning(&packages.Package{
				Name: "baz",
				Syntax: []*ast.File{
					parse(tt.t, tt.expect, `
    type Bar interface {
        Bar()
    }`),
				},
			}, nil),
		))

		found := <-foundCh

		tt.expect(found.Slice()).To(haveLen(1))
		typs := found.Slice()[0].Decls()
		tt.expect(typs).To(haveLen(2))

		spec := typs[1].(*ast.GenDecl).Specs[0].(*ast.TypeSpec)
		tt.expect(spec).To(not(beNil()))
		inter := spec.Type.(*ast.InterfaceType)
		tt.expect(inter.Methods.List).To(haveLen(2))

		bar := inter.Methods.List[0]
		tt.expect(bar.Names[0].String()).To(equal("Bar"))
		b, isFunc := bar.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())
		tt.expect(b.Params.List).To(haveLen(0))
		tt.expect(b.Results).To(beNil())

		foo := inter.Methods.List[1]
		tt.expect(foo.Names[0].String()).To(equal("Foo"))
		f, isFunc := foo.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())
		tt.expect(f.Params.List).To(haveLen(0))
		tt.expect(f.Results).To(beNil())
	})

	o.Spec("AnonymousImportedTypes_Recursion", func(tt testCtx) {
		pers.ConsistentlyReturn(tt.t, tt.dir.PathOutput, "/some/path")
		pers.ConsistentlyReturn(tt.t, tt.dir.PackageOutput, &packages.Package{
			Name: "bar",
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `

    import "some/path/to/foo"

    type Bar interface{
        foo.Foo
        Bar()
    }`),
			},
		})

		pkgName := "foo"
		pkg := &packages.Package{
			Name: pkgName,
			Syntax: []*ast.File{
				parse(tt.t, tt.expect, `
    type Foo interface {
        Foo(func(X) Y) func(Y) X
    }`),
			},
		}
		pers.ConsistentlyReturn(tt.t, tt.dir.ImportOutput, pkg, nil)

		found := typ.Load(false, tt.dir)

		// One call for the initial import, one more for dependency checking
		tt.expect(tt.dir.ImportCalled).To(haveLen(2))
		tt.expect(<-tt.dir.ImportInput.Path).To(equal("some/path/to/foo"))

		tt.expect(found.Slice()).To(haveLen(1))
		typs := found.Slice()[0].Decls()
		tt.expect(typs).To(haveLen(2))

		decl := typs[1]
		tt.expect(decl).To(not(beNil()))
		inter := typeSpec(tt.t, tt.expect, decl).Type.(*ast.InterfaceType)
		tt.expect(inter.Methods.List).To(haveLen(2))

		foo := inter.Methods.List[0]
		tt.expect(foo.Names[0].String()).To(equal("Foo"))
		f, isFunc := foo.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())
		tt.expect(f.Params.List).To(haveLen(1))
		tt.expect(f.Results.List).To(haveLen(1))

		input := f.Params.List[0]
		in, isFunc := input.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())

		expr, isSelector := in.Params.List[0].Type.(*ast.SelectorExpr)
		tt.expect(isSelector).To(beTrue())
		tt.expect(expr.X.(*ast.Ident).String()).To(equal("foo"))
		tt.expect(expr.Sel.String()).To(equal("X"))
		expr, isSelector = in.Results.List[0].Type.(*ast.SelectorExpr)
		tt.expect(isSelector).To(beTrue())
		tt.expect(expr.X.(*ast.Ident).String()).To(equal("foo"))
		tt.expect(expr.Sel.String()).To(equal("Y"))

		output := f.Params.List[0]
		out, isFunc := output.Type.(*ast.FuncType)
		tt.expect(isFunc).To(beTrue())

		expr, isSelector = out.Params.List[0].Type.(*ast.SelectorExpr)
		tt.expect(isSelector).To(beTrue())
		tt.expect(expr.X.(*ast.Ident).String()).To(equal("foo"))
		tt.expect(expr.Sel.String()).To(equal("X"))
		expr, isSelector = out.Results.List[0].Type.(*ast.SelectorExpr)
		tt.expect(isSelector).To(beTrue())
		tt.expect(expr.X.(*ast.Ident).String()).To(equal("foo"))
		tt.expect(expr.Sel.String()).To(equal("Y"))
	})
}

func expectNamesToMatch(t *testing.T, expect expectation, list []ast.Decl, names ...string) {
	t.Helper()
	listNames := make(map[string]struct{}, len(list))
	for _, decl := range list {
		spec := typeSpec(t, expect, decl)
		listNames[spec.Name.String()] = struct{}{}
	}
	expectedNames := make(map[string]struct{}, len(names))
	for _, name := range names {
		expectedNames[name] = struct{}{}
	}
	expect(listNames).To(equal(expectedNames))
}

func typeSpec(t *testing.T, expect expectation, decl ast.Decl) *ast.TypeSpec {
	t.Helper()
	gd, ok := decl.(*ast.GenDecl)
	expect(ok).To(beTrue())
	expect(gd.Specs).To(haveLen(1))
	spec, ok := gd.Specs[0].(*ast.TypeSpec)
	expect(ok).To(beTrue())
	return spec
}

func find(t *testing.T, expect expectation, decls []ast.Decl, name string) *ast.TypeSpec {
	t.Helper()
	for _, decl := range decls {
		gd, ok := decl.(*ast.GenDecl)
		if !ok {
			continue
		}
		if len(gd.Specs) != 1 {
			continue
		}
		typ, ok := gd.Specs[0].(*ast.TypeSpec)
		if !ok {
			continue
		}
		if typ.Name.String() == name {
			return typ
		}
	}
	return nil
}
